package ec.app.sequence;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;

import org.apache.commons.io.FileUtils;
import org.biojava.bio.Annotation;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.io.FastaFormat;
import org.biojava.bio.seq.io.SeqIOTools;
/**
 * This Utility is used to extract Positive and Negative samples from Fasta files which have labels in properties.
 * This is used in Genome Wide data for Splice Sites, like Worm etc where genome wide sequences are given and
 * we need to extract certain amount of positive and negative samples for testing.
 * @author udaykamath
 *
 */
public class FastaParserGenome {
	
	
	
	public static void main(String[] args) {
		
		if(args.length <3){
			System.err.println("FastaParserGenome source maxPositive maxNegative outputFile");
			System.exit(-1);
		}
		try{
		// get positive file name
		String positiveFileName = args[0];
		
		BufferedReader positiveBuffer;
			// read the files
			positiveBuffer = new BufferedReader(
					new FileReader(positiveFileName));

			SequenceIterator positiveFileIterator = (SequenceIterator) SeqIOTools
					.fileToBiojava("fasta", "DNA", positiveBuffer);
			int positiveCount =0;
			int negativeCount =0;
			int maxPositive = new Integer(args[1]);
			int maxNegative = new Integer(args[2]);
			File outputFile = new File(args[3]);
			String sequenceAndLabel ="";
			//find the center
			int center = 398/2;
			int leftWindow = center-60;
			int rightWindow = center+80;
			ArrayList lines = new ArrayList();
			ArrayList shortenedList = new ArrayList();
			ArrayList positiveList = new ArrayList();
			ArrayList negativeList = new ArrayList();
			while(positiveFileIterator.hasNext()){
				Sequence fastaSequence = positiveFileIterator.nextSequence();
				Annotation annotations = fastaSequence.getAnnotation();
				String description = (String)annotations.getProperty(FastaFormat.PROPERTY_DESCRIPTIONLINE);
				//get the start of 
				String label="";
				boolean process = false;
				if(description.contains("label=-1") && negativeCount < maxNegative){
					//negative
					label= " -1";
					negativeCount = negativeCount +1;
					negativeList.add(fastaSequence);
					process = true;
				}
				else if(description.contains("label=1") && (positiveCount < maxPositive)){
					label= " 1";
					positiveCount = positiveCount +1;	
					positiveList.add(fastaSequence);
					process = true;
				} 
				if(process){
					sequenceAndLabel = fastaSequence.subStr(leftWindow, rightWindow);
					if(sequenceAndLabel.trim().length() <1){
						System.out.println("Something wrong here!");
					}
					sequenceAndLabel = sequenceAndLabel + label;
					System.out.println("Current positives: " + positiveCount + " current Negative: " + negativeCount);
					lines.add(sequenceAndLabel);
					shortenedList.add(fastaSequence);
				}
					//if we have reached stop
					if((positiveCount == maxPositive)&& (maxNegative == negativeCount))
							break;
				
			}
			FileUtils.writeLines(outputFile, lines);
			System.out.println("Completed writing fasta file combined");
			//create a fasta file
			FileOutputStream fos = new FileOutputStream("short.fasta");
			//send the sequences
			for(int k=0; k< shortenedList.size(); k++)
				SeqIOTools.writeFasta(fos, (Sequence)shortenedList.get(k));
			fos.flush();
			fos.close();
			System.out.println("Completed writing short.fasta");
			//write to positive and negative file
			//create a fasta file
			FileOutputStream fosPositive = new FileOutputStream(new File("PositiveWorm.fasta"));
			//send the sequences
			for(int k=0; k< positiveList.size(); k++)
				SeqIOTools.writeFasta(fosPositive, (Sequence)positiveList.get(k));
			fosPositive.flush();
			fosPositive.close();
			System.out.println("Completed writing PostiveWorm.fasta");
			
			FileOutputStream fosNegative = new FileOutputStream(new File("NegativeWorm.fasta"));
			//send the sequences
			for(int k=0; k< negativeList.size(); k++)
				SeqIOTools.writeFasta(fosNegative, (Sequence)negativeList.get(k));
			fosNegative.flush();
			fosNegative.close();
			System.out.println("Completed writing NegativeWorm.fasta");
		}catch(Exception e){
			e.printStackTrace();
		}
			
		}
	
	

}
