package ec.app.sequence;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.StringTokenizer;

import org.biojava.bio.seq.DNATools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.io.SeqIOTools;
import org.biojava.bio.symbol.FiniteAlphabet;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.utils.regex.Matcher;
import org.biojava.utils.regex.Pattern;
import org.biojava.utils.regex.PatternFactory;



import ec.EvolutionState;
import ec.Individual;
import ec.Problem;
import ec.gp.GPIndividual;
import ec.gp.GPProblem;
import ec.gp.koza.KozaFitness;
import ec.simple.SimpleFitness;
import ec.simple.SimpleProblemForm;
import ec.util.Parameter;

/**
 * Basic Fitness assessment code. This is meant to be generic for any sequence classification problem.
 * What we do is read the parameters for data file that has positive and negative sequences. Format of file
 * should be SEQUENCE \t label
 * like 
 * ACTTGGCTTGTTGGTTAAGGTTATG.........TAGA 1
 * ACGTTGGTTGAAGGTTAAGGTTATG.........GAGA -1 etc
 * It uses the basic fitness function as 
 * TruePositiveRatio * Math.abs(InformationGain )as fitness.
 * TruePosiiveRatio ~ to Precision value and InformationGain as Discrimination factor.
 * KozaFitness is about mimimization, so we do 1/1+fitness to get the koza fitness.
 * @author udaykamath
 *
 */

public class SequenceClassificationProblem extends GPProblem implements
		SimpleProblemForm {
	protected static double positiveSequences;
	protected static double negativeSequences;
	protected static ArrayList<String> positiveSequencesList = new ArrayList<String>();
	protected static ArrayList<String> negativeSequencesList = new ArrayList<String>();
	private PatternFactory[] factory;
	protected SymbolList currentSequence;
	
	 /** The small deviation allowed in double comparisons. */
	  public static double SMALL = 1e-9;
	  
	  public static double MEDIUM = 1e-5;
	  

	public static final String POSITIVE_FILE_NAME_PARAM = "positivefile";
	/** The natural logarithm of 2 */
	  private static double log2 = Math.log(2);

	
	public PatternFactory getFactory(int thread){
		return factory[thread];
	}
	
	public void setup(final EvolutionState state, final Parameter base) {
		// make sure parents are ok
		super.setup(state, base);
		// get positive file name
		String positiveFileName = state.parameters.getString(base
				.push(POSITIVE_FILE_NAME_PARAM), null);
		BufferedReader reader = null;
		
		try {

			reader = new BufferedReader(new InputStreamReader(
					new FileInputStream(positiveFileName)));

			String totalLine = reader.readLine();
			
			int i = 0;
			while (totalLine != null) {
				StringTokenizer stk = new StringTokenizer(totalLine, " ");
				//int startOfSpace = totalLine.indexOf(" ");
				//String sequence = totalLine.substring(0, startOfSpace);
				//String label = totalLine.substring(startOfSpace + 1, totalLine
					//	.length());
				String label = stk.nextToken();
				String sequence = (String)stk.nextElement();
				if (label.equals("1"))
					positiveSequencesList.add(sequence);
				else 
					negativeSequencesList.add(sequence);
				totalLine = reader.readLine();
				System.out.println("Completed:" + i++);
			}
		}
		catch(Exception e){
			e.printStackTrace();
		}
		FiniteAlphabet iupac = DNATools.getDNA();
		factory = new PatternFactory[4];
		for(int i=0; i<4; i++)
			factory[i] = PatternFactory.makeFactory(iupac);
		
		positiveSequences = this.positiveSequencesList.size();
		negativeSequences = this.negativeSequencesList.size();
	}

	

//	public void evaluate(final EvolutionState state, final Individual ind,
//			final int subpopulation, final int threadnum) {
//		if (!ind.evaluated)  // don't bother reevaluating
//        {
//			int truePositiveCount =0;
//			int trueNegativeCount =0;
//			
//			//iterate positive sequences 
//			for(int i=0; i< positiveSequencesList.size(); i++){
//				String sequence = positiveSequencesList.get(i);
//				try{
//					this.currentSequence = DNATools.createDNA(sequence);
//				}
//				catch(Exception e){
//					e.printStackTrace();
//				}
//				MotifData motifData = new MotifData();
//				//now evaluate the problem on GP
//				((GPIndividual)ind).trees[0].child.eval(
//	                    state,threadnum, motifData,stack,((GPIndividual)ind),this);
//				if(motifData.matches){
//					truePositiveCount = truePositiveCount +1;
//				}
//			}
//			//iterate over negative sequences
//			for(int i=0; i< negativeSequencesList.size(); i++){
//				String sequence = negativeSequencesList.get(i);
//				try{
//					this.currentSequence = DNATools.createDNA(sequence);
//				}
//				catch(Exception e){
//					e.printStackTrace();
//				}
//				MotifData motifData = new MotifData();
//				//now evaluate the problem on GP
//				((GPIndividual)ind).trees[0].child.eval(
//	                    state,threadnum, motifData,stack,((GPIndividual)ind),this);
//				if(motifData.matches){
//					trueNegativeCount = trueNegativeCount +1;
//				}
//			}
//			
//			double precision = 0;
//			double informationGain =0;
//			double combined = SMALL;
//		    precision = truePositiveCount/positiveSequences;
//		    double negative = trueNegativeCount/negativeSequences;
//		    informationGain = Math.abs(precision-negative);			    	
//		    if(informationGain >0 && precision > 0)
//		    	combined = precision * informationGain;
//			
//		
//		
//		
//			//form koza fitness
//			KozaFitness f = ((KozaFitness)ind.fitness);
//			//to avoid ideal individual
//			//if(precision == 1.0)
//			//	precision = 0.9;
//			f.setStandardizedFitness(state,(float)(1.0/combined));
//			//System.out.println("PositiveMatchRatio: " + precision +  " NegativeMatchRatio: " + negative  + " information gain:" + informationGain + " standard koza fitnes:" + f.fitness());
//			//set hits
//			f.hits = positiveSequencesList.size();
//			//set evaluated
//			ind.evaluated = true;   		
//        }
//        
//	}
	
	public void evaluate(final EvolutionState state, final Individual ind,
			final int subpopulation, final int threadnum) {
		if (!ind.evaluated)  // don't bother reevaluating
        {
			int truePositiveCount =0;
			int trueNegativeCount =0;
			
			//iterate positive sequences 
			for(int i=0; i< positiveSequencesList.size(); i++){
				String sequence = positiveSequencesList.get(i);
				try{
					this.currentSequence = DNATools.createDNA(sequence);
				}
				catch(Exception e){
					e.printStackTrace();
				}
				MotifData motifData = new MotifData();
				//now evaluate the problem on GP
				((GPIndividual)ind).trees[0].child.eval(
	                    state,threadnum, motifData,stack,((GPIndividual)ind),this);
				if(motifData.matches){
					truePositiveCount = truePositiveCount +1;
				}
			}
			//iterate over negative sequences
			for(int i=0; i< negativeSequencesList.size(); i++){
				String sequence = negativeSequencesList.get(i);
				try{
					this.currentSequence = DNATools.createDNA(sequence);
				}
				catch(Exception e){
					e.printStackTrace();
				}
				MotifData motifData = new MotifData();
				//now evaluate the problem on GP
				((GPIndividual)ind).trees[0].child.eval(
	                    state,threadnum, motifData,stack,((GPIndividual)ind),this);
				if(motifData.matches){
					trueNegativeCount = trueNegativeCount +1;
				}
			}
			
			ind.printIndividualForHumans(state,0,1000);
			StringWriter writer = new StringWriter();
			PrintWriter printWriter = new PrintWriter(writer);
			ind.printIndividual(state, printWriter);
			printWriter.flush();
			writer.flush();
			String representation = writer.toString();
			System.out.println("True Positive count: " + truePositiveCount +  " and True Negative Count: " + trueNegativeCount +" for " + representation);
			double precision = 0;
			precision = truePositiveCount/positiveSequences;
		    double negative = trueNegativeCount/negativeSequences;
		    double informationGain = Math.abs(precision-negative);		
			double combined = SMALL;
			boolean dontWrite = false;
			if(!representation.contains("matches")  && !representation.contains("present")){
				dontWrite = true;
			}						
			//fitness is precision
			//else if((representation.contains("matches") && truePositiveCount >0) || (representation.contains("present") && truePositiveCount >0)){
			else{
//			    precision = truePositiveCount/positiveSequences;
//			    double[][] counts = new double[2][2];
//			    counts[0][0] = positiveSequences-truePositiveCount;
//			    counts[0][1] = negativeSequences-trueNegativeCount;
//			    counts[1][0] = truePositiveCount;
//			    counts[1][1]=  trueNegativeCount;
//			    //caclulate information gain
//			    informationGain = Math.abs(entropyConditionedOnColumns(counts)- entropyConditionedOnRows(counts));
			    System.out.println("InformationGain: " + informationGain);
			    if(informationGain >0 && precision > 0)
			    	combined =  precision * informationGain;
			}
			
			
			
			//form koza fitness
			KozaFitness f = ((KozaFitness)ind.fitness);
			//to avoid ideal individual
			//if(precision == 1.0)
			//	precision = 0.9;
			f.setStandardizedFitness(state,(float)(1.0/combined)); 
			//set hits
			f.hits = positiveSequencesList.size();
			//set evaluated
			ind.evaluated = true;   			
//			if(truePositiveCount > 3 && !dontWrite && combined > SMALL)
//				FeatureCollector.writeFeature((GPIndividual)ind, state);
			
			System.out.println("Finished evaulating fitness:" + combined);
        }
	}
	
	 /**
	   * Computes conditional entropy of the rows given
	   * the columns.
	   *
	   * @param matrix the contingency table
	   * @return the conditional entropy of the rows given the columns
	   */
	  public static double entropyConditionedOnColumns(double[][] matrix) {
	    
	    double returnValue = 0, sumForColumn, total = 0;

	    for (int j = 0; j < matrix[0].length; j++) {
	      sumForColumn = 0;
	      for (int i = 0; i < matrix.length; i++) {
		returnValue = returnValue + lnFunc(matrix[i][j]);
		sumForColumn += matrix[i][j];
	      }
	      returnValue = returnValue - lnFunc(sumForColumn);
	      total += sumForColumn;
	    }
	    if (eq(total, 0)) {
	      return 0;
	    }
	    return -returnValue / (total * log2);
	  }

	  /**
	   * Computes conditional entropy of the columns given
	   * the rows.
	   *
	   * @param matrix the contingency table
	   * @return the conditional entropy of the columns given the rows
	   */
	  public static double entropyConditionedOnRows(double[][] matrix) {
	    
	    double returnValue = 0, sumForRow, total = 0;

	    for (int i = 0; i < matrix.length; i++) {
	      sumForRow = 0;
	      for (int j = 0; j < matrix[0].length; j++) {
		returnValue = returnValue + lnFunc(matrix[i][j]);
		sumForRow += matrix[i][j];
	      }
	      returnValue = returnValue - lnFunc(sumForRow);
	      total += sumForRow;
	    }
	    if (eq(total, 0)) {
	      return 0;
	    }
	    return -returnValue / (total * log2);
	  }
	  
	  /**
	   * Tests if a is equal to b.
	   *
	   * @param a a double
	   * @param b a double
	   */
	  public static /*@pure@*/ boolean eq(double a, double b){
	    
	    return (a - b < SMALL) && (b - a < SMALL); 
	  }
	  
	  /**
	   * Help method for computing entropy.
	   */
	  private static double lnFunc(double num){
	    
	    // Constant hard coded for efficiency reasons
	    if (num < 1e-6) {
	      return 0;
	    } else {
	      return num * Math.log(num);
	    }
	  }
	
	public SymbolList getCurrentSequence(){
		return this.currentSequence;
	}
	
	
}
